from collections import defaultdict
from os.path import join
import os
import _pickle as cPickle
import common.seg as seg
from itertools import repeat
import common.L1L2_cells_ids as layers

def getSurvivingBetween(tss, linss, t1, t2):
    linsL = filterLinssBetween(linss, (t1, t2))

    cids = list()
    
    for c in tss[t1]:
        if tSearchN(linsL, c.cid, len(linsL), 0):
            cids.append(c.cid)

    return cids

def filterSurvivingUntil(tss, linss, t2):
    for t, ts in tss.items():
        cids = getSurvivingBetween(tss, linss, t, t2)
        filterT(ts, cids)

    return

def fst(x): return x[0]

def snd(x): return x[1]

def filterT(ts, cids):
    scids = set(cids)
    ts.cells = dict([(c.cid, c) for c in ts if c.cid in scids])

    neighsL1 = dict()
    for cid, ds in ts.neighs.items():
        if cid in scids:
            neighsL1[cid] = dict([(n, d) for n, d in ds.items()
                                  if n in scids])

    ts.neighs = neighsL1
        
def filterL1(tss):
    L1_cids = {10: layers.L1_10h,
               40: layers.L1_40h,
               96: layers.L1_96h,
               120: layers.L1_120h,
               132: layers.L1_132h}
    
    for t, ts in tss.items():
        filterT(ts, L1_cids[t])

    return


def filterL1L2(tss):
    L1_cids = {10: layers.L1_10h,
               40: layers.L1_40h,
               96: layers.L1_96h,
               120: layers.L1_120h,
               132: layers.L1_132h}

    L2_cids = {10: layers.L2_10h,
               40: layers.L2_40h,
               96: layers.L2_96h,
               120: layers.L2_120h,
               132: layers.L2_132h}

    for t, ts in tss.items():
        filterT(ts, L1_cids[t] + L2_cids[t])

    return


def getTs(linFn):
    fn, _ = os.path.splitext(linFn)
    (_, t1, _, t2) = fn.split("_")

    return (int(t1[:-1]), int(t2[:-1]))


def readLin(linFn):
    with open(linFn, 'rb') as linF:
        lin = cPickle.load(linF, encoding='latin1')

    return lin, getTs(linFn)


def readLins(linDataLoc):
    lins = [readLin(join(linDataLoc, f)) for f in os.listdir(linDataLoc)
            if os.path.isfile(join(linDataLoc, f))]

    return lins


def getLinMap(lin):
    cMap = defaultdict(list)
    for c1, c2 in lin[0]:
        cMap[c2].append(c1)
        
    return cMap


def mkLinsSeries(lins):
    return [(ts, getLinMap(lin))
            for lin, ts in sorted(lins, key=lambda lt: lt[1][0])]


def after(t, lt):
    ts, _ = lt
    (t1, t2) = ts
    
    return t >= t1


def before(t, lt):
    ts, _ = lt
    (t1, t2) = ts
    
    return t <= t2


def between(linSeries, t1, t2):
    afters = [i for i, lt in enumerate(linSeries) if after(t1, lt)]
    befores = [i for i, lt in enumerate(linSeries) if before(t2, lt)]

    s = afters[-1]
    e = befores[0]

    return [(ts, lin) for ts, lin in linSeries[s:(e+1)]]


def mergeLins(lin1, lin2):
    mLin = dict()
    
    for m, ds in lin1.items():
        mLin[m] = sum([lin2[d] for d in ds], [])

    return mLin


def mergeLinss1(lins, acc):
    if not lins: return acc

    return mergeLinss1(lins[1:], mergeLins(acc, lins[0]))


def mergeLinss(lins):
    if len(lins) > 2: return mergeLinss1(lins[2:], mergeLins(lins[0], lins[1]))
    elif len(lins) == 2: return mergeLins(lins[0], lins[1])
    elif len(lins) == 1: return lins[0]
    else: return []

    
######################################################
def concat(xss):
    return sum(xss, [])


def tSearchN(mss, r, n, c):
    if c == n: return [r]
    else:
        return concat([tSearchN(mss, d, n, c+1) for d in mss[c].get(r, [])])

    
def tSearch1(mss, r):
    try:
        return mss[0][r]
    except KeyError:
        return []

    
def tSearch2(mss, r):
    try:
        return sum([mss[1][d] for d in mss[0][r]], [])
    except KeyError:
        print(r)
        return []

    
def cvol(c):
    if c: return c.vol
    else: return 0.0

    
def sumVol(cs, f=cvol):
    return sum([cvol(c) for c in cs])


def mergeCellVolsRefN(tss, mss):
    volss = []

    for i, ts in enumerate(tss):
        volss.append(dict())
        for c in tss[0]:
            volss[i][c.cid] = sum(cvol(ts.cells.get(ci, None))
                                  for ci in tSearchN(mss, c.cid, i, 0))

    return volss


def const1(c):
    if c: return 1
    else: return 0


def mergeCellNsRefN(tss, linss):
    attrss = []

    for i, ts in enumerate(tss):
        attrss.append(dict())
        for c in tss[0]:
            attrss[i][c.cid] = sum(const1(ts.cells.get(ci, None))
                                   for ci in tSearchN(linss, c.cid, i, 0))

    return  attrss


def mergeCellVolsRef3(tss, mss):
    (ts1, ts2, ts3) = tss
    
    vols1 = dict()
    vols2 = dict()
    vols3 = dict()
    
    for c in ts1:
        vols1[c.cid] = c.vol
        vols2[c.cid] = sumVol([ts2.cells[i] for i in tSearch1(mss, c.cid)])
        vols3[c.cid] = sumVol([ts3.cells[i] for i in tSearch2(mss, c.cid)])

    return vols1, vols2, vols3


def mergeCellVolsRef2(tss, mss):
    (ts1, ts2) = tss
    
    vols1 = dict()
    vols2 = dict()
    
    for c in ts1:
        print(c.cid)
        vols1[c.cid] = c.vol
        vols2[c.cid] = sumVol([ts2.cells[i] for i in tSearch1(mss, c.cid)])

    return vols1, vols2


def divOr0(n1, n2):
    if n2 == 0: return 0
    else: return n1 / n2

    
def growthRates3(volss, m12):
    vols1, vols2, vols3 = volss
    gr2 = dict()
    for i, vol in vols1.items():
        gr12 = divOr0((vols2[i] - vols1[i]), vols1[i])
        gr23 = divOr0((vols3[i] - vols2[i]), vols2[i])
        gr2[i] = 0.5 * (gr12 + gr23)
        
    gr22 = [zip(m12.get(i, []), repeat(gr, len(m12.get(i, []))))
            for i, gr in gr2.items()]
    
    return dict(sum(gr22, []))


def growthRatesL2(volss):
    vols1, vols2 = volss
    gr1 = dict()
    for i, vol in vols1.items():
        gr1[i] = (vols2[i] - vols1[i]) / vols1[i]

    return gr1


def growthRatesR2(volss, m12):
    vols1, vols2 = volss
    gr2 = dict()
    for i, vol in vols1.items():
        gr2[i] = (vols2[i] - vols1[i]) / vols1[i]

    gr22 = [zip(m12.get(i, []), repeat(gr, len(m12.get(i, []))))
            for i, gr in gr2.items()]
    
    return dict(sum(gr22, []))


def mkSeries1(dataDir):
    d = "organism"                                                                                 
    dExprs = "geneExpression/new_patterns"
    linDataLoc = "tracking"
    

    timepoints = [10, 40, 96, 120, 132]
    lins = mkLinsSeries(readLins(join(dataDir, linDataLoc)))
     
    tss = dict()
    linss = dict()

    for t1, t2 in zip(timepoints, timepoints[1:]):
        tss[t1] = readDataT1(join(dataDir, d), join(dataDir, dExprs), t1)
        tss[t2] = readDataT1(join(dataDir, d), join(dataDir, dExprs), t2)

        linss[(t1, t2)] = mergeLinss(list(map(lambda x: x[1],
                                              between(lins, t1, t2))))

    return tss, linss

def mkSeriesIm0(dataDir):
    d = "yr01"
    linDataLoc = "yr01/tracking"
     
    timepoints = [0, 10, 40, 96, 120, 132]
    lins = mkLinsSeries(readLins(join(dataDir, linDataLoc)))
     
    tss = dict()
    linss = dict()

    for t in timepoints:
        tss[t] = readDataT1Im(join(dataDir, d), t)
         
    for t1, t2 in zip(timepoints, timepoints[1:]):
        linss[(t1, t2)] = mergeLinss(map(lambda x: x[1],
                                          between(lins, t1, t2)))

    return tss, linss


def mkGrowthRates():
    tss, linss = mkSeries()
    grs = dict()

    grs[10] = growthRatesL2(mergeCellVolsRef2((tss[10], tss[40]),
                                              [linss[(10, 40)]]))
    grs[40] = growthRates3(mergeCellVolsRef3((tss[10], tss[40], tss[96]),
                                             [linss[(10, 40)], linss[(40, 96)]]),
                           linss[(10, 40)])
    grs[96] = growthRates3(mergeCellVolsRef3((tss[40], tss[96], tss[120]),
                                             [linss[(40, 96)], linss[(96, 120)]]),
                           linss[(40, 96)])
    grs[120] = growthRates3(mergeCellVolsRef3((tss[96], tss[120], tss[132]),
                                              [linss[(96, 120)], linss[(120, 132)]]),
                            linss[(96, 120)])
    grs[132] = growthRatesR2(mergeCellVolsRef2((tss[120], tss[132]),
                                               [linss[(120, 132)]]),
                             linss[(120, 132)])

    return grs

def readGeomImT(dataDir, fid, t):
    ressFn = "yr" + fid + "/resolutions.txt"
    segFn = ("yr" +
             fid +
             "/segmentations/YR_" +
             fid + "_" + str(t) +
             "h_segmented.tif")
    
    ress = readRess(join(dataDir, ressFn))
    img = readImage(join(dataDir, segFn))
    res = ress[t]
    
    ts = Tissue.fromImage(data=img, res=res)
    for c in ts:
        c.swapZX()
        
    return ts

def mkSeriesGeomIm(dataDir, fid):
    ressFn = "yr" + fid + "/resolutions.txt"
    linDataLoc = "yr" + fid + "/trackingFiles/"
    ress = readRess(join(dataDir, ressFn))
    tpoints = sorted(ress.keys())
    
    lins = mkLinsSeries(readLins(join(dataDir, linDataLoc)))
    tss = dict()
    linss = dict()
    
    for t in tpoints:
        tss[t] = readGeomImT(dataDir, fid, t)
        
    for t1, t2 in zip(tpoints, tpoints[1:]):
        linss[(t1, t2)] = mergeLinss(map(lambda x:
                                         x[1], between(lins, t1, t2)))
        
    return tss, linss


###################################
def invertMap(ln):
    iLin = [list(zip(ds, repeat(m, len(ds)))) for m, ds in ln.items()]

    return dict(sum(iLin, []))

def updateCells(ts, ts1, iLn):
    for c1 in ts1:
        try:
            c = ts.cells[iLn[c1.cid]]
        except KeyError:
            print(c1.cid)
            c = Cell(0, Vec3(0, 0, 0), 0, dict([(g, False) for g in geneNms]))
        c1.exprs = dict([(g, c.exprs[g]) for g, v in c1.exprs.items()])

    return

def prev(t):
    if t == 40: return 10
    elif t == 96: return 40
    elif t == 120: return 96
    elif t == 132: return 120
    else: return None

def succ(t):
    if t==10: return 40
    elif t==40: return 96
    elif t==96: return 120
    elif t==120: return 132
    else: return None

def updateTs(tss, linss, tpoints):
    #evolve patterns based on lineage
    tssLin = {}
    for t in tpoints[1:]:
        tssLin[(prev(t), t)] = tss[t].updateExprsTs(tss[prev(t)],
                                                    invertMap(linss[(prev(t), t)]))

    return tssLin

def updateTsBF(tss, fs):
    return dict((t, tss[t].updateExprsBF(fs[t])) for t in tss.keys())

def toNewmanFn(fn):
    (fn, ext) = os.path.splitext(fn)
    return fn + "000" + ".tif"


def tFn(t, ref=""):
    return "ts_t" + str(t) + ref + ".txt"


def snd(x):
    return x[1]

def betweenTs(t, bounds):
    ts, te = bounds
    return t >= ts and t <= te

def filterLinssBetween(linss, bounds):
    t1, t2 = bounds
    linssL = sorted([(ts, ln) for ts, ln in linss.items() 
                     if betweenTs(ts[0], (t1, t2)) and
                        betweenTs(ts[1], (t1, t2))], key=lambda p: p[0][0])
    
    return list(map(snd, linssL))

def filterTssBetween(tss, tbounds):
    t1, t2 = tbounds
        
    tssL = sorted([(t, ts) for t, ts in tss.items() if betweenTs(t, (t1, t2))], key=fst)
    
    return list(map(snd, tssL))

def readDataT1(d, dExprs, t):
    from os.path import join
    
    geomF = "_segmented_tvformat_0_volume_position.txt"
    neighF = "_segmented_tvformat_0_neighbors_py.txt"
    exprF = "h.txt"

    fnGeom = join(d, "t" + str(t) + geomF)
    fnNeigh = join(d, "t" + str(t) + neighF)
    fnExpr = join(dExprs, "t_" + str(t) + exprF)
    
    gexprs = seg.readExprs(fnExpr)
    ts = seg.STissue.fromTV(fnGeom, fnNeigh, gexprs, seg.geneNms)

    return ts

def readDataT1Im(d, dExprs, t):
    from os.path import join

    exprF = "h.txt"
    fnRes = join(d, "resolutions.txt")
    fnSeg = join(d, "segmentations/t"+str(t)+"_segmented.tif")
    fnExpr = join(dExprs, "t_" + str(t) + exprF)
    
    
    data, _ = seg.readImage(fnSeg)
    ress = seg.readRess(fnRes)
    res = ress[t]

    gexprs = seg.readExprs(fnExpr)
    ts = seg.STissue.fromImage(data, res, gexprs, seg.geneNms)

    for c in ts:
        c.swapZX()

    return ts

